/* Fractal Image Compression Demonstration */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <alloc.h>

#define IMAGE argv[1]
#define IFS argv[2]

get_image(FILE *ifs_file, unsigned int *assumed_width_of_image, unsigned int *assumed_height_of_image, unsigned int *width_of_image, unsigned int *height_of_image, char *argv[]);
unsigned char domain_block(unsigned long int domx, unsigned char domain[4][4], unsigned int assumed_width_of_image, unsigned int width_of_image, unsigned int height_of_image);
unsigned char range_block(unsigned long int ranx, unsigned char ran[8][8], unsigned int assumed_width_of_image, unsigned int width_of_image, unsigned int height_of_image);
symmetry(unsigned char ss, unsigned char dom[4][4], unsigned char d1[4][4]);
bestfit(unsigned long int ranx, unsigned char s, unsigned char d1[4][4], unsigned char ran[8][8], unsigned long int *bestranx, unsigned char *bestsym, unsigned char *lowest_error);
write_to_file(unsigned long int bestranx, unsigned char bestsym, FILE *ifs_file, unsigned int assumed_width_of_image, char *argv[]);
calculate_time(time_t start_time);

unsigned int *image;

main(int argc, char *argv[])
{
	unsigned char dom[4][4], d1[4][4], ran[8][8], lowest_error, sym, bestsym, domain_whites, range_whites;
	unsigned int assumed_width_of_image, assumed_height_of_image, width_of_image, height_of_image;
	unsigned long int count, domx, domy, ranx, rany, bestranx;
	time_t start_time;
	FILE *ifs_file;

	start_time = time(NULL);
	if (argc != 3) {
		printf("\nUse the following format:\n\ncompress [image_file] [ifs_file]\n\n");
		exit(1);
	}
	if ((ifs_file = fopen(IFS, "wb")) == NULL) {
		fprintf(stderr, "\nError opening file %s\n", IFS);
		exit(1);
	}
	get_image(ifs_file, &assumed_width_of_image, &assumed_height_of_image, &width_of_image, &height_of_image, argv);
	printf("\n\nCompressing %s", IMAGE);
	count = 0;
	for (domy = 0; domy < (assumed_width_of_image/4) * assumed_height_of_image; domy += assumed_width_of_image)
		for (domx = domy; domx < domy + assumed_width_of_image/4; domx++) {
			if (count++ % 200 == 0)
				printf(".");
			lowest_error = 64;
			bestsym = 0;
			domain_whites = domain_block(domx, dom, assumed_width_of_image, width_of_image, height_of_image);
			if (domain_whites == 16) {
				bestranx = 0;
				bestsym = 8;
				write_to_file(bestranx, bestsym, ifs_file, assumed_width_of_image, argv);
				continue;
			}
			for (rany = 0; rany < (assumed_width_of_image/4) * (assumed_height_of_image - 4); rany += assumed_width_of_image) {
				for (ranx = rany; ranx < rany + (assumed_width_of_image/4) - 1; ranx++) {
					range_whites = range_block(ranx, ran, assumed_width_of_image, width_of_image, height_of_image);
					if (range_whites == 64)
						continue;
					if (abs(4 * domain_whites - range_whites) > lowest_error)
						continue;
					for (sym = 0; sym < 8; sym++) {
						symmetry(sym, dom, d1);
						bestfit(ranx, sym, d1, ran, &bestranx, &bestsym, &lowest_error);
						if (lowest_error == 0)
							break;
					}
					if (lowest_error == 0)
						break;
				}
				if (lowest_error == 0)
					break;
			}
			write_to_file(bestranx, bestsym, ifs_file, assumed_width_of_image, argv);
		}
	free(image);
	fclose(ifs_file);
	calculate_time(start_time);
	return 0;
}

get_image(FILE *ifs_file, unsigned int *assumed_width_of_image, unsigned int *assumed_height_of_image, unsigned int *width_of_image, unsigned int *height_of_image, char *argv[])
{
	FILE *image_file;
	unsigned int buf[24], *header, size_of_header, extra_height;
	unsigned long size_of_file, size_of_image;

	if ((image_file = fopen(IMAGE, "rb")) == NULL) {
		fprintf(stderr, "\nCannot open file %s\n", IMAGE);
		exit(1);
	}
	if (fread(buf, sizeof(unsigned int), 24, image_file) != 24) {
		fprintf(stderr, "\nError reading file %s\n", IMAGE);
		exit(1);
	}
	if (buf[0] != 19778) {
		printf("%s is not a .BMP file", IMAGE);
		exit(0);
	}
	if (buf[23] != 2) {
		printf("%s is not a black and white image", IMAGE);
		exit(0);
	}
	size_of_file = buf[2] * 65536 + buf[1];
	size_of_header = buf[5];
	size_of_image = size_of_file - size_of_header;
	*width_of_image = buf[9];
	if (*width_of_image % 32 != 0)
		*assumed_width_of_image = *width_of_image + 32 - (*width_of_image % 32);
	else *assumed_width_of_image = *width_of_image;
	*height_of_image = buf[11];
	if (*height_of_image%4 !=0)
		*assumed_height_of_image = *height_of_image + 4 - (*height_of_image % 4);
	else *assumed_height_of_image = *height_of_image;
	extra_height = ((*assumed_height_of_image - *height_of_image) * *assumed_width_of_image)/8;
	if (fseek(image_file, 0, SEEK_SET) != 0) {
		fprintf(stderr, "\nError with file %s\n", IMAGE);
		exit(1);
	}
	if ((header = (unsigned int *) calloc(size_of_header/2, sizeof(unsigned int))) == NULL) {
		fprintf(stderr, "\nUnable to allocate sufficient memory for header\n");
		exit(1);
	}
	if (fread(header, sizeof(unsigned int), size_of_header/2, image_file) != size_of_header/2) {
		fprintf(stderr, "\nCannot read header from %s\n", IMAGE);
		exit(1);
	}
	fwrite(header, sizeof(unsigned int), size_of_header/2, ifs_file);
	if ((image = (unsigned int *) calloc((size_of_image + extra_height) / 2, sizeof(unsigned int))) == NULL) {
		fprintf(stderr, "\nUnable to assign enough memory for image\n");
		exit(1);
	}
	if (fread(image, sizeof(unsigned int), size_of_image/2, image_file) != size_of_image/2) {
		fprintf(stderr, "\nError reading file %s\n", IMAGE);
		exit(1);
	}
	fclose(image_file);
	return 0;
}

unsigned char domain_block(unsigned long int dx, unsigned char domain[4][4], unsigned int assumed_width_of_image, unsigned int width_of_image, unsigned int height_of_image)
{
	unsigned char i, j, whites = 0;
	unsigned long xi, xa;

	for (xi = dx, i = 0; i < 4 ; xi += assumed_width_of_image / 4, i++) {
		xa = (xi % 4 < 2) ? xi + 2 : xi - 2;
		for (j = 0; j < 4; j++) {
			if (((4 * xi) / assumed_width_of_image < height_of_image) && ((4 * xi + j) % assumed_width_of_image < width_of_image))
				domain[i][j] = 1 & (image[xa/4] >> (15 - ((xa % 4) * 4 + j)));
			else domain[i][j] = 1;
			if (domain[i][j] == 1)
				whites++;
		}
	}
	return whites;
}

unsigned char range_block(unsigned long int ranx, unsigned char ran[8][8], unsigned int assumed_width_of_image, unsigned int width_of_image, unsigned int height_of_image)
{
	unsigned char i, j, d[4][4], ran_whites = 0;

	ran_whites += domain_block(ranx, d, assumed_width_of_image, width_of_image, height_of_image);
	for (i = 0; i < 4; i++)
		for (j = 0; j < 4; j++)
			ran[i][j] = d[i][j];
	ran_whites += domain_block(ranx + 1, d, assumed_width_of_image, width_of_image, height_of_image);
	for (i = 0; i < 4; i++)
		for (j = 4; j < 8; j++)
			ran[i][j] = d[i][j-4];
	ran_whites += domain_block(ranx + assumed_width_of_image, d, assumed_width_of_image, width_of_image, height_of_image);
	for (i = 4; i < 8; i++)
		for (j = 0; j < 4; j++)
			ran[i][j] = d[i-4][j];
	ran_whites += domain_block(ranx+1+assumed_width_of_image, d, assumed_width_of_image, width_of_image, height_of_image);
	for (i = 4; i < 8; i++)
		for (j = 4; j < 8; j++)
			ran[i][j] = d[i-4][j-4];
	return ran_whites;
}

symmetry(unsigned char sym, unsigned char dom[4][4], unsigned char d1[4][4])
{
	unsigned char i, j;

	for (i = 0; i < 4; i++)
		for (j = 0; j < 4; j++)
			switch (sym) {
			case 0:
				d1[i][j] = dom[i][j];
				break;
			case 1:
				d1[i][j] = dom[3-i][j];
				break;
			case 2:
				d1[i][j] = dom[i][3 - j];
				break;
			case 3:
				d1[i][j] = dom[3 - i][3 - j];
				break;
			case 4:
				d1[i][j] = dom[j][i];
				break;
			case 5:
				d1[i][j] = dom[j][3 - i];
				break;
			case 6:
				d1[i][j] = dom[3 - j][i];
				break;
			case 7:
				d1[i][j] = dom[3 - j][3 - i];
				break;
			}
	return 0;
}

bestfit(unsigned long int ranx, unsigned char s, unsigned char d1[4][4], unsigned char ran[8][8], unsigned long int *bestranx, unsigned char *bestsym, unsigned char *lowest_error)
{
	unsigned char i, j, current_error = 0;

	for (i = 0; i < 8 ; i++)
		for (j = 0; j < 8; j++) {
			if (ran[i][j] != d1[i/2][j/2])
				current_error++;
		}
	if (current_error < *lowest_error) {
		*lowest_error = current_error;
		*bestranx = ranx;
		*bestsym = s;
	}
	return 0;
}

write_to_file(unsigned long int bestranx, unsigned char bestsym, FILE *ifs_file, unsigned int assumed_width_of_image, char *argv[])
{
	unsigned int abestranx;

	if (putc(bestsym, ifs_file) == EOF) {
		fprintf(stderr, "\nError writing to file %s\n", IFS);
		exit(1);
	}
	if (bestsym != 8) {
		abestranx = bestranx - (3 * assumed_width_of_image / 4 + 1) * (bestranx / assumed_width_of_image);
		if (putw(abestranx, ifs_file) == EOF) {
			fprintf(stderr, "\nError writing to file %s\n", IFS);
			exit(1);
		}
	}
	return 0;
}

calculate_time(time_t start_time)
{
	unsigned int hours, mins, secs;
	unsigned long tim;
	time_t finish_time;

	finish_time = time(NULL);
	tim = difftime(finish_time, start_time);
	hours = tim / 3600;
	mins = (tim - 3600 * (tim / 3600)) / 60;
	secs = tim % 60;
	printf("\n\nCompression time: ");
	if (hours == 1)
		printf("%u hour ", hours);
	else if (hours > 1)
		printf("%u hours ", hours);
	if (mins == 1)
		printf("%u minute ", mins);
	else if(mins > 1)
		printf("%u minutes ", mins);
	if (secs == 1)
		printf("%u second", secs);
	else
		printf("%u seconds", secs);
	printf("\n\n");
	return 0;
}
