/* Fractal Image Decompression Demonstration */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <alloc.h>

#define IFS argv[1]
#define DECOMIMAGE argv[2]
#define NO_OF_ITERATIONS 3

load_header(unsigned int *assumed_width_of_image, unsigned int *height_of_image, unsigned int *assumed_height_of_image, unsigned int *size_of_header, unsigned long *size_of_image, FILE *ifs_file, FILE *decomp_file, char *argv[]);
initialize_screens(unsigned long size_of_image, unsigned int assumed_width_of_image, unsigned int height_of_image, unsigned int assumed_height_of_image, unsigned int *extra_height);
read_ifs_file(unsigned int assumed_width_of_image, unsigned long int *bestranx, unsigned char *bests, FILE *ifs_file, char *argv[]);
get_domain_block(unsigned long int dx, unsigned char domain[4][4], unsigned int assumed_width_of_image);
get_range_block(unsigned long int ranx, unsigned int assumed_width_of_image, unsigned char ran[8][8]);
replace_domain_block(unsigned int s, unsigned char dom[4][4], unsigned char ran[8][8]);
new_domain_block(unsigned long int dx, unsigned int assumed_width_of_image, unsigned char dom[4][4]);
copy_domain_screen_to_range_screen(unsigned long int size_of_image, unsigned int extra_height);
copy_domain_screen_to_decomp_file(unsigned long size_of_image, FILE *decomp_file);
calculate_time(time_t start_time);

unsigned int *domain_screen, *range_screen;

main(int argc, char *argv[])
{
	unsigned char iter, i, j, bests, dom[4][4], ran[8][8];
	unsigned int assumed_width_of_image, height_of_image, assumed_height_of_image, size_of_header, extra_height;
	unsigned long count, size_of_image, domx, domy, bestranx;
	time_t start_time;
	FILE *ifs_file, *decomp_file;

	start_time = time(NULL);

	if (argc != 3) {
		printf("\nUse the following format:\n\ndecompress [ifs_file] [image_file]\n\n");
		exit(1);
	}
	if ((ifs_file = fopen(IFS, "rb")) == NULL) {
		printf("\nCannot open file %s\n", IFS);
		exit(1);
	}
	if ((decomp_file = fopen(DECOMIMAGE, "wb")) == NULL) {
		fprintf(stderr, "\nCannot open file %s\n", DECOMIMAGE);
		exit(1);
	}
	load_header(&assumed_width_of_image, &height_of_image, &assumed_height_of_image, &size_of_header, &size_of_image, ifs_file, decomp_file, argv);
	initialize_screens(size_of_image, assumed_width_of_image, height_of_image, assumed_height_of_image, &extra_height);
	printf("\n\nDecompressing %s", IFS);
	count = 0;
	for(iter = 0; iter < NO_OF_ITERATIONS; iter++) {
		fseek(ifs_file, size_of_header, SEEK_SET);
		for (domy = 0; domy < (assumed_width_of_image/4) * assumed_height_of_image; domy += assumed_width_of_image)
			for (domx = domy; domx < domy + assumed_width_of_image/4; domx++) {
				if (count++ % (200 * NO_OF_ITERATIONS) == 0)
					printf(".");
				read_ifs_file(assumed_width_of_image, &bestranx, &bests, ifs_file, argv);
				if (bests != 8) {
					get_range_block(bestranx, assumed_width_of_image, ran);
					replace_domain_block(bests, dom, ran);
				}
				else for (i = 0; i < 4; i++)
						for (j = 0; j < 4; j++)
							dom[i][j] = 1;
				new_domain_block(domx, assumed_width_of_image, dom);
			}
		if (iter < NO_OF_ITERATIONS - 1) {
			copy_domain_screen_to_range_screen(size_of_image, extra_height);
		}
	}
	fclose(ifs_file);
	free(range_screen);
	copy_domain_screen_to_decomp_file(size_of_image, decomp_file);
	fclose(decomp_file);
	free(domain_screen);
	calculate_time(start_time);
	return 0;
}

load_header(unsigned int *assumed_width_of_image, unsigned int *height_of_image, unsigned int *assumed_height_of_image, unsigned int *size_of_header, unsigned long *size_of_image, FILE *ifs_file, FILE *decomp_file, char *argv[])
{
	unsigned int buf[24], *header, width_of_image;
	unsigned long size_of_file;

	if (fread(buf, sizeof(unsigned int), 24, ifs_file) != 24) {
		fprintf(stderr, "\nError reading file %s\n", IFS);
		exit(1);
	}
	if (buf[0] != 19778) {
		printf("%s is not an .IFS file", IFS);
		exit(0);
	}
	if (buf[23] != 2) {
		printf("%s is not a black and white image", IFS);
		exit(0);
	}
	size_of_file = buf[2] * 65536 + buf[1];
	*size_of_header = buf[5];
	*size_of_image = size_of_file - *size_of_header;
	width_of_image = buf[9];
	if (width_of_image % 32 != 0)
		*assumed_width_of_image = width_of_image + 32 - (width_of_image % 32);
	else *assumed_width_of_image = width_of_image;
	*height_of_image = buf[11];
	if (*height_of_image % 4 != 0)
		*assumed_height_of_image = *height_of_image + 4 - (*height_of_image % 4);
	else *assumed_height_of_image = *height_of_image;
	fseek(ifs_file, 0, SEEK_SET);
	if ((header = (unsigned int *) calloc(*size_of_header/2, sizeof(unsigned int))) == NULL) {
		fprintf(stderr, "\nUnable to allocate sufficient memory for header\n");
		exit(1);
	}
	if (fread(header, sizeof(unsigned int), *size_of_header/2, ifs_file) != *size_of_header/2) {
		fprintf(stderr, "\nCannot read header from %s\n", IFS);
		exit(1);
	}
	if (fwrite(header, 1, *size_of_header, decomp_file) != *size_of_header) {
		printf("\nCannot write header\n");
		exit(1);
	}
	free(header);
	return 0;
}

initialize_screens(unsigned long size_of_image, unsigned int assumed_width_of_image, unsigned int height_of_image, unsigned int assumed_height_of_image, unsigned int *extra_height)
{
	*extra_height = ((assumed_height_of_image - height_of_image) * assumed_width_of_image)/8;
	if ((domain_screen = (unsigned int *) calloc((size_of_image + *extra_height)/2, sizeof(unsigned int))) == NULL) {
		fprintf(stderr, "\nUnable to allocate sufficient memory for domain screen\n");
		exit(1);
	}
	if ((range_screen = (unsigned int *) calloc((size_of_image + *extra_height)/2, sizeof(unsigned int))) == NULL) {
		fprintf(stderr, "\nUnable to allocate sufficient memory for range screen\n");
		exit(1);
	}
	return 0;
}

read_ifs_file(unsigned int assumed_width_of_image, unsigned long int *bestranx, unsigned char *bests, FILE *ifs_file, char *argv[])
{
	unsigned int abestranx;

	if (fread(bests, sizeof(unsigned char), 1, ifs_file) != 1) {
		fprintf(stderr, "\nError reading file %s\n", IFS);
		exit(1);
	}
	if (*bests != 8) {
		abestranx = getw(ifs_file);
		if (ferror(ifs_file)) {
			fprintf(stderr, "\nError reading file %s\n", IFS);
			exit (1);
		}
		*bestranx = assumed_width_of_image * ((4 * (abestranx)) / (assumed_width_of_image - 4)) + (abestranx % ((assumed_width_of_image - 4) / 4));
	}
	return 0;
}

get_domain_block(unsigned long int dx, unsigned char domain[4][4], unsigned int assumed_width_of_image)
{
	unsigned char i, j;
	unsigned long int xi;

	dx = (dx % 4 < 2) ? dx + 2 : dx - 2;
	for (xi = dx, i = 0; i < 4 ; xi += assumed_width_of_image / 4, i++)
		for (j = 0; j < 4; j++)
			domain[i][j] = 1 & (range_screen[xi/4] >> (15 - ((xi % 4) * 4 + j)));
	return 0;
}

get_range_block(unsigned long int ranx, unsigned int assumed_width_of_image, unsigned char ran[8][8])
{
	unsigned char i, j, d[4][4];

	get_domain_block(ranx, d, assumed_width_of_image);
	for (i = 0; i < 4; i++)
		for (j = 0; j < 4; j++)
			ran[i][j] = d[i][j];
	get_domain_block(ranx + 1, d, assumed_width_of_image);
	for (i = 0; i < 4; i++)
		for (j = 4; j < 8; j++)
			ran[i][j] = d[i][j-4];
	get_domain_block(ranx + assumed_width_of_image, d, assumed_width_of_image);
	for (i = 4; i < 8; i++)
		for (j = 0; j < 4; j++)
			ran[i][j] = d[i-4][j];
	get_domain_block(ranx+1+assumed_width_of_image, d, assumed_width_of_image);
	for (i = 4; i < 8; i++)
		for (j = 4; j < 8; j++)
			ran[i][j] = d[i-4][j-4];
	return 0;
}

replace_domain_block(unsigned int s, unsigned char dom[4][4], unsigned char ran[8][8])
{
	unsigned char i, j;

	for (i = 0; i < 4; i++)
		for (j = 0; j < 4; j++) {
			switch (s) {
			case 0:
				dom[i][j] = ran[2*i][2*j] + ran[2*i][2*j+1] + ran[2*i+1][2*j] + ran[2*i+1][2*j+1];
				break;
			case 1:
				dom[i][j] = ran[7-2*i][2*j] + ran[7-2*i][2*j+1] + ran[7-(2*i+1)][2*j] + ran[7-(2*i+1)][2*j+1];
				break;
			case 2:
				dom[i][j] = ran[2*i][7-2*j] + ran[2*i][7-(2*j+1)] + ran[2*i+1][7-2*j] + ran[2*i+1][7-(2*j+1)];
				break;
			case 3:
				dom[i][j] = ran[7-2*i][7-2*j] + ran[7-2*i][7-(2*j+1)] + ran[7-(2*i+1)][7-2*j] + ran[7-(2*i+1)][7-(2*j+1)];
				break;
			case 4:
				dom[i][j] = ran[2*j][2*i] + ran[2*j][2*i+1] + ran[2*j+1][2*i]+ran[2*j+1][2*i+1];
				break;
			case 5:
				dom[i][j] = ran[2*j][7-2*i] + ran[2*j][7-(2*i+1)] + ran[2*j+1][7-2*i] + ran[2*j+1][7-(2*i+1)];
				break;
			case 6:
				dom[i][j] = ran[7-2*j][2*i] + ran[7-2*j][2*i+1] + ran[7-(2*j+1)][2*i] + ran[7-(2*j+1)][2*i+1];
				break;
			case 7:
				dom[i][j] = ran[7-2*j][7-2*i] + ran[7-2*j][7-(2*i+1)] + ran[7-(2*j+1)][7-2*i] + ran[7-(2*j+1)][7-(2*i+1)];
				break;
			}
	if (dom[i][j] < 2)
		dom[i][j] = 0;
	else if (dom[i][j] > 2)
		dom[i][j] = 1;
	else
		dom[i][j] = rand() % 2;
	}
	return 0;
}

new_domain_block(unsigned long int dx, unsigned int assumed_width_of_image, unsigned char dom[4][4])
{
	unsigned char i;
	unsigned int xi;

	union {
		unsigned int word;
		struct {
			unsigned int bit0 : 1;
			unsigned int bit1 : 1;
			unsigned int bit2 : 1;
			unsigned int bit3 : 1;
			unsigned int bit4 : 1;
			unsigned int bit5 : 1;
			unsigned int bit6 : 1;
			unsigned int bit7 : 1;
			unsigned int bit8 : 1;
			unsigned int bit9 : 1;
			unsigned int bit10 : 1;
			unsigned int bit11 : 1;
			unsigned int bit12 : 1;
			unsigned int bit13 : 1;
			unsigned int bit14 : 1;
			unsigned int bit15 : 1;
		} bits;
	} temp;

	for (xi = dx, i = 0; i < 4 ; xi += assumed_width_of_image/4, i++) {
		temp.word = domain_screen[xi/4];
		if (xi % 4 == 0) {
				temp.bits.bit4 = dom[i][3];
				temp.bits.bit5 = dom[i][2];
				temp.bits.bit6 = dom[i][1];
				temp.bits.bit7 = dom[i][0];
		}
		if (xi % 4 == 1) {
				temp.bits.bit0 = dom[i][3];
				temp.bits.bit1 = dom[i][2];
				temp.bits.bit2 = dom[i][1];
				temp.bits.bit3 = dom[i][0];
		}
		if (xi % 4 == 2) {
				temp.bits.bit12 = dom[i][3];
				temp.bits.bit13 = dom[i][2];
				temp.bits.bit14 = dom[i][1];
				temp.bits.bit15 = dom[i][0];
		}
		if (xi % 4 == 3) {
				temp.bits.bit8 = dom[i][3];
				temp.bits.bit9 = dom[i][2];
				temp.bits.bit10 = dom[i][1];
				temp.bits.bit11 = dom[i][0];
		}
		domain_screen[xi/4] = temp.word;
	}
	return 0;
}

copy_domain_screen_to_range_screen(unsigned long int size_of_image, unsigned int extra_height)
{
	unsigned long int i;

	for (i = 0; i < (size_of_image + extra_height) / 2; i++)
		range_screen[i] = domain_screen[i];
	return 0;
}

copy_domain_screen_to_decomp_file(unsigned long size_of_image, FILE *decomp_file)
{
	fwrite(domain_screen, sizeof(unsigned int), size_of_image/2, decomp_file);
	return 0;
}

calculate_time(time_t start_time)
{
	unsigned int hours, mins, secs;
	unsigned long tim;
	time_t finish_time;

	finish_time = time(NULL);
	tim = difftime(finish_time, start_time);
	hours = tim / 3600;
	mins = (tim - 3600 * (tim / 3600)) / 60;
	secs = tim % 60;
	printf("\n\nDecompression time: ");
	if (hours == 1)
		printf("%u hour ", hours);
	else if (hours > 1)
		printf("%u hours ", hours);
	if (mins == 1)
		printf("%u minute ", mins);
	else if(mins > 1)
		printf("%u minutes ", mins);
	if (secs == 1)
		printf("%u second", secs);
	else
		printf("%u seconds", secs);
	printf("\n\n");
	return 0;
}
